use std::collections::{BTreeMap, VecDeque};
use std::io::{self, Write};


pub unsafe fn raw_alloc<T>() -> *mut T {
    unsafe { std::alloc::alloc_zeroed(std::alloc::Layout::new::<T>()) as *mut T }
}

pub unsafe fn raw_free<T>(ptr: *mut T) {
    unsafe { std::alloc::dealloc(ptr as *mut u8, std::alloc::Layout::new::<T>()) }
}

#[derive(Debug)]
pub struct Student {
    id: u32,
    sis_id: u32,
}

impl Student {
    fn new(id: u32, sis_id: u32) -> *mut Student {
        let result: *mut Student = unsafe {raw_alloc::<Student>()};
        let result_ref: &mut Student = unsafe{&mut *result};
        result_ref.id = id;
        result_ref.sis_id = sis_id;
        return result;
    }

    fn show(&self) {
        print!("Student#{} [SIS ID#{}]", self.id, self.sis_id);
    }
}

#[derive(Debug)]
pub struct Submission {
    grade: Option<u32>,
    student1: *mut Student,
    student2: *mut Student,
}

impl Submission {
    pub fn new(student_ptr1: *mut Student, student_ptr2: *mut Student) -> *mut Submission {
        let result: *mut Submission = unsafe{raw_alloc::<Submission>()};
        let result_ref: &mut Submission = unsafe{&mut *result};
        result_ref.grade = None;
        result_ref.student1 = student_ptr1;
        result_ref.student2 = student_ptr2;
        return result;
    }

    pub fn show_grade(&self) {
        match self.grade {
            None => print!("[ungraded]"),
            Some(grade) => print!("grade {}", grade),
        }
    }

    pub fn show(&self) {
        let student_ref1: &Student = unsafe{&*self.student1};
        let student_ref2: &Student = unsafe{&*self.student2};
        print!("Submission for ");
        student_ref1.show();
        print!(" and ");
        student_ref2.show();
        print!(" ");
        self.show_grade();
    }
}

#[derive(Debug)]
pub struct Class {
    students: BTreeMap<u32, *mut Student>,
    graded_submissions: VecDeque<*mut Submission>,
    ungraded_submissions: VecDeque<*mut Submission>,
}

fn show_submission_deque(submissions: &VecDeque<*mut Submission>) {
    for submission_ptr_ref in submissions.iter() {
        let submission_ptr: *const Submission = *submission_ptr_ref;
        let submission_ref: &Submission = unsafe{&*submission_ptr};
        submission_ref.show();
        print!("\n");
    }
}

impl Class {
    pub fn new() -> Class {
        Class {
            students: BTreeMap::new(),
            graded_submissions: VecDeque::new(),
            ungraded_submissions: VecDeque::new(),
        }
    }

    pub fn show(&self) {
        for (_k, v_ref) in self.students.iter() {
            // (k, v_ref) is a (&u32, &*mut Student)
            let student_ptr: *mut Student = *v_ref;
            let student_ref: &Student = unsafe{&*student_ptr};
            student_ref.show();
            print!("\n");
        }
        println!("Graded submissions:");
        show_submission_deque(&self.graded_submissions);
        println!("Ungraded submission [next-to-grade last]:");
        show_submission_deque(&self.ungraded_submissions);
    }
}

pub fn show_help() {
    print!(concat!(
           "Usage:\n",
           "  show\n",
           "  new-student ID SIS-ID\n",
           "  new-submission STUDENT-ID1 STUDENT-ID2\n",
           "  show-next\n",
           "  grade-next GRADE\n",
           "  drop-student ID\n",
           "  exit\n"));
}

pub fn run_command(class: &mut Class, command: &str, args: Vec<u32>) {
    match command {
        "help" => show_help(),
        "show" => {
            if args.len() != 0 {
                show_help();
            } else {
                class.show();
            }
        }
        "new-student" => {
            if args.len() != 2 {
                show_help();
            } else {
                class.students.insert(
                    args[0],
                    Student::new(args[0], args[1])
                );
                println!("Added student #{}", class.students.len());
            }
        }
        "new-submission" => {
            if args.len() != 2 {
                show_help();
            } else {
                match (class.students.get(&args[0]), class.students.get(&args[1])) {
                    // (Some(&*mut Student), Some(&*mut Student))
                    (Some(student_ptr_ref1), Some(student_ptr_ref2)) => {
                        let student_ptr1: *mut Student = *student_ptr_ref1;
                        let student_ptr2: *mut Student = *student_ptr_ref2;
                        let submission = Submission::new(
                            student_ptr1, student_ptr2
                        );
                        class.ungraded_submissions.push_front(submission);
                        println!("Created submission");
                    },
                    _ => {
                        eprintln!("Could not find students")
                    }
                }
            }
        },
        "show-next" => {
            if args.len() != 0 {
                show_help();
            } else {
                match class.ungraded_submissions.back() {
                    Some(submission_ptr_ref) => {
                        // submission_ptr_ref is &*mut Submission
                        let submission_ptr: *const Submission = *submission_ptr_ref;
                        let submission: &Submission = unsafe{&*submission_ptr};
                        submission.show();
                        print!("\n");
                    },
                    None => {
                        eprintln!("No submissions to grade.");
                    },
                }
            }
        },
        "grade-next" => {
            if args.len() != 1 {
                show_help();
            } else {
                match class.ungraded_submissions.pop_back() {
                    Some(submission_ptr) => {
                        let submission: &mut Submission = unsafe{&mut *submission_ptr};
                        submission.grade = Some(args[0]);
                        class.graded_submissions.push_back(submission_ptr);
                    },
                    None => {
                        eprintln!("No submissions to grade.");
                    },
                }
            }
        },
        "drop-student" => {
            if args.len() != 1 {
                show_help();
            } else {
                match class.students.remove(&args[0]) {
                    Some(student_ptr) => {
                        unsafe {raw_free(student_ptr)};
                        println!("Removed student '{}'", args[0]);
                    },
                    None => {
                        eprintln!("No such student '{}'", args[0])
                    }
                }
            }
        }
        _ => {
            show_help();
        }
    }
}

fn main() {
    let mut class = Class::new();
    loop {
        let mut line = String::new();
        print!("> ");
        io::stdout().flush().expect("I/O error");
        io::stdin().read_line(&mut line).expect("error reading input");
        let parts: Vec<&str> = line.split_whitespace().collect();
        if parts.len() == 0 {
            show_help();
            continue
        }
        if parts[0] == "exit" {
            break
        }
        let mut args: Vec<u32> = Vec::new();
        for arg_str in &parts[1..] {
            match arg_str.parse::<u32>() {
                Ok(value) => args.push(value),
                _ => {
                    eprintln!("'{}' is not a non-negative number.", arg_str);
                    break;
                }
            }
        }
        run_command(&mut class, parts[0], args);
    }
}
